import nest
import os
import numpy as np
import pandas as pd
import scanpy as sc
from tqdm import tqdm
from sklearn.metrics import jaccard_score
from collections import defaultdict
import pickle

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib import rc

# rc('font', **{'family': 'serif', 'serif': ['Arial'], 'size': 5})
import seaborn as sns


def generate_scores(adata):
    # Run a bunch of HMRF partitions for various values of number of communities
    # k_range = [5, 10, 15, 20, 25]
    k_range = range(2, 31)
    targets = [0, 2, 4, 5, 6, 8, 16, 20]
    methods = ["hmrf", "bayesspace", "spagcn"]
    scores = {k: defaultdict(list) for k in methods}
    for k in tqdm(k_range):
        spagcn = nest.methods.SpaGCN(regions=k)
        spagcn.fit(adata, verbose=False)

        bayesspace = nest.methods.BayesSpace(regions=k)
        bayesspace.fit(adata)
        # exit(0)

        hmrf = nest.hmrf.HMRFSegmentationModel(
            adata, regions=k, eps=180, label_name="class_hmrf", beta=1
        )
        hmrf.fit(max_iterations=200, update_labels=True, verbose=False)

        # find the region with the greatest overlap with each target CH
        partitions = {
            "hmrf": adata.obs["class_hmrf"],
            "bayesspace": adata.obs["class_bayesspace"],
            "spagcn": adata.obs["class_spagcn"],
        }
        for method, partition in partitions.items():
            for target in targets:
                target_membership = pd.notnull(
                    adata.obs[f"hotspots_multi_{target}"]
                ).to_numpy()
                max_score = -1
                max_partition = None
                for val in np.unique(partition):
                    score = jaccard_score(target_membership, partition == val)
                    if score > max_score:
                        max_score = score
                        max_partition = partition
                scores[method][f"ch_{target}"].append((max_score, max_partition))
        with open("test.pkl", "wb") as f:
            pickle.dump(scores, f)


if __name__ == "__main__":
    height = 1.2
    dataset = "V1_Mouse_Brain_Sagittal_Anterior"
    cache_dir = os.path.expanduser(f"data/{dataset}")
    image_save_dir = os.path.expanduser(f"images/{dataset}/")
    if not os.path.exists(image_save_dir):
        os.makedirs(image_save_dir)
    nest.plot.set_dataset_plot_parameters(dataset)
    rc("font", **{"family": "serif", "serif": ["Arial"], "size": 5})
    adata = nest.read(os.path.join(cache_dir, "adata"))
    # adata.layers['exp'] = adata.X.expm1()

    # generate_scores(adata)
    # exit(0)

    with open("data/test.pkl", "rb") as f:
        scores = pickle.load(f)

    chs_to_plot = [0, 2, 6, 8, 16, 20]

    fig, axs = plt.subplots(1, 6, figsize=(6.5, 1.5))
    for idx, ch in enumerate(chs_to_plot):
        ax = axs[idx]
        d = {
            method: [v[0] for v in data[f"ch_{ch}"]] for method, data in scores.items()
        }
        df = pd.DataFrame(d, index=list(range(2, 31)))
        if idx == 5:
            legend = True
        else:
            legend = False
        sns.scatterplot(data=df, ax=ax, legend=legend, markers=True)
        ax.set_ylim([0.0, 1.0])
        ax.set_title(f"CH {ch}")
    plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)
    plt.tight_layout()
    fig.savefig(
        os.path.join(image_save_dir, f"comparison.pdf"),
        dpi=300,
        transparent=True,
        bbox_inches="tight",
    )
    plt.show()



import warnings
import scipy.sparse
import scanpy as sc
from scipy.sparse import coo_array
import leidenalg as la
from scipy.spatial import KDTree
from tqdm import tqdm
import igraph as ig
import pandas as pd

import numpy as np

from sklearn.preprocessing import normalize
from scipy.spatial import Delaunay

from nest.utility import get_neighbor_adjacency


def compute_spatial_transport_matrices(
    adata, secreted_std, secreted_threshold_cutoff, contact_threshold=None, z_key=None
):
    """
    Return the spatial prior belief of possibility of interaction between every possible pair of
    cells for both secreted signaling and cell-cell contact type interactions
    """

    ncells = adata.shape[0]

    # gaussian diffusion kernel - threshold is two standard deviations
    sigma = secreted_std

    def weight_fn(x):
        return np.exp(-0.5 * ((x / (2 * sigma)) ** 2))

    coords = adata.obsm["spatial"]
    if z_key is not None:
        z = np.array(adata.obs[z_key]).reshape(-1, 1)
        coords = np.concatenate([coords, z], axis=1)
    transport_secreted = get_neighbor_adjacency(
        coords, eps=sigma * secreted_threshold_cutoff, weight=weight_fn
    )
    if z_key is None:
        # delauney for direct contact
        tri = Delaunay(adata.obsm["spatial"])
        indptr, indices = tri.vertex_neighbor_vertices
        data = np.ones(shape=indices.shape)
        transport_contact = scipy.sparse.csr_matrix(
            (data, indices, indptr), shape=[ncells, ncells]
        )

        # filter out edges in delauney triangulation that are too long
        if contact_threshold is not None:
            transport_contact = transport_contact.multiply(
                get_neighbor_adjacency(coords, eps=contact_threshold)
            )

        # normalize across number of neighbors
        transport_contact = normalize(transport_contact, norm="l1", axis=1)
        adata.obsp["transport_contact"] = transport_contact

    # transport_secreted = normalize(transport_secreted, norm='l1', axis=1)

    adata.obsp["transport_secreted"] = transport_secreted


def segmentation_3d(
    adata, z_key="Bregma", output_key="segmentation_3d", inplace=True, n_pca=8
):
    if not inplace:
        adata = adata.copy()
    kv = 20
    k_inter = 5
    gamma = 1
    omega = 0.0001
    alpha = 0.2

    layer_offset = 0
    layer_offset_list = [0]

    within_adj_list = []
    inter_layer_edges = []

    # PCA reduction over all layers
    sc.tl.pca(adata, n_comps=n_pca)

    z_values = np.unique(adata.obs[z_key])

    for layer in tqdm(range(12)):
        adata_sub_1 = adata[adata.obs["Bregma"] == z_values[layer], :]
        X_pca = adata_sub_1.obsm["X_pca"]
        xs = adata_sub_1.obsm["spatial"]
        kd_1 = KDTree(xs)
        dist_1, nn_1 = kd_1.query(x=xs, k=range(2, kv + 2))
        dist_weights = np.exp(-(dist_1 / 0.04))

        kd_pca = KDTree(X_pca)
        dist_pca, nn_pca = kd_pca.query(x=X_pca, k=range(2, kv + 2))
        exp_weights = np.exp(
            -np.linalg.norm(X_pca.reshape((-1, 1, n_pca)) - X_pca[nn_1, :], axis=-1) / 2
        )

        # compute within_layer adjacency
        i = np.ravel([(i,) * kv for i in range(len(xs))])
        j = np.ravel(nn_1)
        data = np.ravel(dist_weights)
        within_adj_dist = coo_array(
            (data, (i + layer_offset, j + layer_offset)), shape=(len(adata), len(adata))
        ).tocsr()

        i = np.ravel([(i,) * kv for i in range(len(xs))])
        j = np.ravel(nn_pca)
        data = np.ravel(exp_weights)
        within_adj_expr = coo_array(
            (data, (i + layer_offset, j + layer_offset)), shape=(len(adata), len(adata))
        ).tocsr()

        try:
            within_adj = (
                alpha * within_adj_dist + (1 - alpha) * within_adj_expr
            ).tocoo()
        except ValueError:
            print(within_adj_dist.shape, within_adj_expr.shape)

        within_adj_list.append(within_adj)

        # add to between layer adjacency
        if layer < 11:
            adata_sub_2 = adata[adata.obs["Bregma"] == z_values[layer + 1], :]
            xt = adata_sub_2.obsm["spatial"]
            kd_2 = KDTree(xt)

            dist_12, nn_12 = kd_1.query(x=xt, k=k_inter)
            dist_21, nn_21 = kd_2.query(x=xs, k=k_inter)

            # C = cdist(adata_sub_1.obsm['spatial'], adata_sub_2.obsm['spatial'], metric='sqeuclidean')
            A = np.zeros((len(xs), len(xt)))

            for i, row in enumerate(nn_21):
                for j in row:
                    weight = 1 / k_inter
                    inter_layer_edges.append(
                        (i + layer_offset, j + layer_offset + len(xs), omega * weight)
                    )
                    A[i, j] = 1
            for j, row in enumerate(nn_12):
                for i in row:
                    weight = 1 / k_inter
                    inter_layer_edges.append(
                        (i + layer_offset, j + layer_offset + len(xs), omega * weight)
                    )
                    A[i, j] = 1

        layer_offset += len(xs)
        layer_offset_list.append(layer_offset)

    partitions = []
    for adj in within_adj_list:
        adj.resize((layer_offset, layer_offset))
        g = ig.Graph.Weighted_Adjacency(adj)
        g.vs["node_size"] = 1
        partitions.append(
            la.RBConfigurationVertexPartition(
                g, weights="weight", resolution_parameter=gamma
            )
        )

    inter_layer_edges_arr = np.array(inter_layer_edges).T
    data = inter_layer_edges_arr[2, :]
    i = inter_layer_edges_arr[0, :].astype(np.int_)
    j = inter_layer_edges_arr[1, :].astype(np.int_)
    inter_layer_adj = coo_array((data, (i, j)))
    inter_layer_adj.resize((layer_offset, layer_offset))
    inter_g = ig.Graph.Weighted_Adjacency(inter_layer_adj)
    inter_g.vs["node_size"] = 0
    partitions.append(
        la.CPMVertexPartition(
            inter_g, weights="weight", resolution_parameter=0.0, node_sizes="node_size"
        )
    )

    optimiser = la.Optimiser()

    max_iters = 3
    for _ in tqdm(range(max_iters)):
        diff = optimiser.optimise_partition_multiplex(partitions)
        if diff == 0:
            break

    full_membership = np.array(partitions[0].membership)

    adata.obs[output_key] = pd.Categorical(full_membership)

    if not inplace:
        return adata



import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import os


def filter_param(df, param):
    filter_set = {0, 1, 2, 3} - {param}
    return df[~df["varied"].isin(filter_set)]


if __name__ == "__main__":
    multi_dfs = []
    for spatial_genes in [24, 32, 64, 256, 512]:
        dfs = [
            pd.read_csv(
                f"data/synthetic_hierarchy/results_{spatial_genes}_{trial}.csv",
                index_col=0,
            )
            for trial in range(10)
        ]
        df = pd.concat(dfs)
        df["spatial_genes"] = spatial_genes
        multi_dfs.append(df)

    combined_df = pd.concat(multi_dfs)

    combined_df = combined_df[combined_df["eps"] > 0.02]

    combined_df["spatial_genes"] = pd.Categorical(combined_df["spatial_genes"])

    combined_df.to_csv("data/synthetic_hierarchy/results.csv")

    sns.set(style="whitegrid", font_scale=0.75)

    fig, axs = plt.subplots(1, 4, figsize=[6, 2])

    sns.lineplot(
        filter_param(combined_df, 0),
        x="eps",
        y="score",
        hue="spatial_genes",
        ax=axs[0],
        legend=None,
    )
    axs[0].set_xlim(left=0.05, right=0.2)
    axs[0].set_xlabel("ε")
    axs[0].set_ylabel("Jaccard similarity")

    sns.lineplot(
        filter_param(combined_df, 1),
        x="density",
        y="score",
        hue="spatial_genes",
        ax=axs[1],
        legend=None,
    )

    axs[1].set_ylabel("")

    sns.lineplot(
        filter_param(combined_df, 2),
        x="threshold",
        y="score",
        hue="spatial_genes",
        ax=axs[2],
        legend=None,
    )

    axs[2].set_ylabel("")

    sns.lineplot(
        filter_param(combined_df, 3),
        x="resolution",
        y="score",
        hue="spatial_genes",
        ax=axs[3],
    )

    axs[3].set_ylabel("")
    axs[3].legend(title="spatial genes")
    sns.move_legend(axs[3], "center left", bbox_to_anchor=(1, 0.5))

    for k in range(4):
        if k != 0:
            axs[k].set_ylabel("")
            axs[k].axes.yaxis.set_ticklabels([])
        axs[k].set_ylim([0, 1])

    fig.savefig(
        os.path.expanduser("data/synthetic_hierarchy/plot.pdf"),
        dpi=300,
        transparent=True,
        bbox_inches="tight",
    )



from goatools.base import download_go_basic_obo

obo_fname = download_go_basic_obo()
from goatools.base import download_ncbi_associations

fin_gene2go = download_ncbi_associations()
from goatools.obo_parser import GODag

obodag = GODag("go-basic.obo")

from goatools.anno.genetogo_reader import Gene2GoReader

# Read NCBI's gene2go. Store annotations in a list of namedtuples
objanno = Gene2GoReader(fin_gene2go, taxids=[10090])
ns2assoc = objanno.get_ns2assc()

for nspc, id2gos in ns2assoc.items():
    print("{NS} {N:,} annotated mouse genes".format(NS=nspc, N=len(id2gos)))
from goatools.cli.ncbi_gene_results_to_python import ncbi_tsv_to_py

ncbi_tsv = "gene_result.txt"
output_py = "genes_ncbi_10090_proteincoding.py"
ncbi_tsv_to_py(ncbi_tsv, output_py)
# Compute GO annotations for hotspots
from goatools.obo_parser import GODag
from goatools.anno.genetogo_reader import Gene2GoReader
from genes_ncbi_10090_proteincoding import GENEID2NT as GeneID2nt_mus
from goatools.base import download_ncbi_associations
from goatools.goea.go_enrichment_ns import GOEnrichmentStudyNS

fin_gene2go = download_ncbi_associations()
obodag = GODag("go-basic.obo")
objanno = Gene2GoReader(fin_gene2go, taxids=[10090])
ns2assoc = objanno.get_ns2assc()

goeaobj = GOEnrichmentStudyNS(
    GeneID2nt_mus.keys(),  # List of mouse protein-coding genes
    ns2assoc,  # geneid/GO associations
    obodag,  # Ontologies
    propagate_counts=False,
    alpha=0.05,  # default significance cut-off
    methods=["fdr_bh"],
)  # defult multipletest correction method

symbol_to_id = {v[5]: v[2] for v in GeneID2nt_mus.values()}
id_to_symbol = {v: k for k, v in symbol_to_id.items()}


def convert_symbol_to_id(gene_symbols):
    # TODO: consider adding warning for genes that were not encoded to id
    return [symbol_to_id[v] for v in gene_symbols if v in symbol_to_id]


def convert_id_to_symbol(ids):
    return [id_to_symbol[v] for v in ids]


def go_analysis(gene_ids, save_name, id):
    goea_results_all = goeaobj.run_study(gene_ids, prt=None)
    goea_results_sig = [r for r in goea_results_all if r.p_fdr_bh < 0.05]
    print(
        "\tSignificant results: {E} enriched, {P} purified".format(
            E=sum(1 for r in goea_results_sig if r.enrichment == "e"),
            P=sum(1 for r in goea_results_sig if r.enrichment == "p"),
        )
    )
    save_dir = f"go_results/{save_name}"
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    goeaobj.wr_xlsx(os.path.join(save_dir, f"coex_{id}.xlsx"), goea_results_sig)
    return goea_results_all, goea_results_sig
